class Error(Exception):
    pass


class InvalidHeaderMagic(Error):
    pass


class InvalidBPB(Error):
    pass


class BadClusterError(Error):
    pass


class FreeClusterError(Error):
    pass


class EmptyDirectoryError(Error):
    pass


class LastEmptyDirectoryError(Error):
    pass


class InvalidDirectoryError(Error):
    pass


class FileNotFoundError(Error):
    pass


class NotADirectoryError(Error):
    pass



from dissect import cstruct

# https://ogris.de/fatrepair/fat.c
c_fat_def = """
#define ATTR_READ_ONLY 0x01
#define ATTR_HIDDEN    0x02
#define ATTR_SYSTEM    0x04
#define ATTR_VOLUME_ID 0x08
#define ATTR_DIRECTORY 0x10
#define ATTR_ARCHIVE   0x20
#define ATTR_LONG_NAME (ATTR_READ_ONLY | ATTR_HIDDEN | ATTR_SYSTEM | ATTR_VOLUME_ID)
#define ATTR_LONG_NAME_MASK (ATTR_READ_ONLY | ATTR_HIDDEN | ATTR_SYSTEM | ATTR_VOLUME_ID | ATTR_DIRECTORY | ATTR_ARCHIVE)

#define LAST_LONG_ENTRY 0x40

enum Fattype {
    FATunknown,
    FAT12,
    FAT16,
    FAT32
};

struct Bpb {
    uint8_t  BS_jmpBoot[3];    /* jump instruction to boot code */
    uint8_t  BS_OEMName[8];    /* "MSWIN4.1" */
    uint16_t BPB_BytsPerSec;   /* bytes per sector (512, 1k, 2k, 4k) */
    uint8_t  BPB_SecPerClus;   /* sectors per cluster (2^n, 0<=n<=7) */
    uint16_t BPB_RsvdSecCnt;   /* number of reserved sectors */
    uint8_t  BPB_NumFATs;      /* count of FATs on the volume (usually 2) */
    uint16_t BPB_RootEntCnt;   /* count of root directory entries (0 if FAT32) */
    uint16_t BPB_TotSec16;     /* total count of sectors (0 if FAT32) */
    uint8_t  BPB_Media;        /* media type, usally 0xf8 */
    uint16_t BPB_FATSz16;      /* sectors occupied by one fat (FAT12 / FAT16) */
    uint16_t BPB_SecPerTrk;    /* sectors per track for Int 0x13 */
    uint16_t BPB_NumHeads;     /* numbers of heads for Int 0x13 */
    uint32_t BPB_HiddSec;      /* count of sectors preceding the partition */
    uint32_t BPB_TotSec32;     /* total count of all sectors of the volume */
};

struct Bpb16 {
    uint8_t  BS_DrvNum;        /* Int 0x13 drive number, eg. 0x80 */
    uint8_t  BS_Reserved1;     /* reserved for WinNT (usually 0) */
    uint8_t  BS_BootSig;       /* extended boot signature (0x29) */
    uint32_t BS_VolID;         /* volume serial number (date + time) */
    uint8_t  BS_VolLab[11];    /* volume label as stored in the root directory */
    uint8_t  BS_FilSysType[8]; /* informational! */
};

struct Bpb32 {
    uint32_t BPB_FATSz32;      /* sectors occupied by one fat (FAT32) */
    uint16_t BPB_ExtFlags;     /* FAT mirrored? */
    uint16_t BPB_FSVer;        /* version number of FAT filesystem type */
    uint32_t BPB_RootClus;     /* cluster number of first cluster of root dir */
    uint16_t BPB_FSInfo;       /* sector number of FSINFO (usually 1) */
    uint16_t BPB_BkBootSec;    /* sector number of copy of boot sector */
    uint8_t  BPB_Reserved[12]; /* reserved for future use */

    uint8_t  BS_DrvNum;        /* Int 0x13 drive number, eg. 0x80 */
    uint8_t  BS_Reserved1;     /* reserved for WinNT (usually 0) */
    uint8_t  BS_BootSig;       /* extended boot signature (0x29) */
    uint32_t BS_VolID;         /* volume serial number (date + time) */
    uint8_t  BS_VolLab[11];    /* volume label as stored in the root directory */
    uint8_t  BS_FilSysType[8]; /* informational! */
};

struct Dirent {
    uint8_t  DIR_Name[11];
    uint8_t  DIR_Attr;
    uint8_t  DIR_NTRes;
    uint8_t  DIR_CrtTimeTenth;
    uint16_t DIR_CrtTime;
    uint16_t DIR_CrtDate;
    uint16_t DIR_LstAccDate;
    uint16_t DIR_FstClusHI;
    uint16_t DIR_WrtTime;
    uint16_t DIR_WrtDate;
    uint16_t DIR_FstClusLO;
    uint32_t DIR_FileSize;
};

struct Ldirent {
    uint8_t  LDIR_Ord;
    uint8_t  LDIR_Name1[10];
    uint8_t  LDIR_Attr;
    uint8_t  LDIR_Type;
    uint8_t  LDIR_Chksum;
    uint8_t  LDIR_Name2[12];
    uint16_t LDIR_FstClusLO;
    uint8_t  LDIR_Name3[4];
};
"""  # noqa: E501

c_fat = cstruct.cstruct()
c_fat.load(c_fat_def)

Fattype = c_fat.Fattype

VALID_BPB_MEDIA = {0xF0, 0xF8, 0xF9, 0xFA, 0xFB, 0xFC, 0xFD, 0xFE, 0xFF}

DATA_CLUSTER_MIN = 0x2
DATA_CLUSTER_MAX = 0x0FFFFFEF
END_OF_CLUSTER_MIN = 0x0FFFFFF8
END_OF_CLUSTER_MAX = 0x0FFFFFFF

FAT12_EOC = 0xFF0
BAD_CLUSTER = 0x0FFFFFF7
FREE_CLUSTER = 0x0



import os

import pytest


def absolute_path(filename):
    return os.path.join(os.path.dirname(__file__), filename)


@pytest.fixture
def exfat_simple():
    name = "data/exfat.bin"
    with open(absolute_path(name), "rb") as f:
        yield f


@pytest.fixture
def fat12():
    name = "data/fat12.bin"
    with open(absolute_path(name), "rb") as f:
        yield f


@pytest.fixture
def fat16():
    name = "data/fat16.bin"
    with open(absolute_path(name), "rb") as f:
        yield f


@pytest.fixture
def fat32():
    name = "data/fat32.bin"
    with open(absolute_path(name), "rb") as f:
        yield f



from dissect.fat import fat


def test_fat12(fat12):
    volume_label = "VOLLAB1"

    fs = fat.FATFS(fat12)

    assert fs.fat.bits_per_entry == 12

    assert fs.volume_label == volume_label
    assert fs.volume_id == "e038bb7c"
    assert fs.cluster_size == 512

    verify_fs_content(fs, volume_label)


def test_fat16(fat16):
    volume_label = "LABFAT16"

    fs = fat.FATFS(fat16)

    assert fs.fat.bits_per_entry == 16

    assert fs.volume_label == volume_label
    assert fs.volume_id == "88fa453f"
    assert fs.cluster_size == 512

    verify_fs_content(fs, volume_label)


def test_fat32(fat32):
    volume_label = "LABFAT32"

    fs = fat.FATFS(fat32)

    assert fs.fat.bits_per_entry == 32

    assert fs.volume_label == volume_label
    assert fs.volume_id == "4368dbb7"
    assert fs.cluster_size == 512

    verify_fs_content(fs, volume_label)


def verify_fs_content(fs, volume_label):
    entries_map = {e.name: e for e in fs.root.iterdir()}

    assert set(entries_map.keys()) == {volume_label, "file1.txt", "file2.txt", "subdir1"}

    file1 = entries_map["file1.txt"]
    assert not file1.is_directory()
    assert file1.size == 20

    file2 = entries_map["file2.txt"]
    assert not file2.is_directory()
    assert file2.size == 23

    subdir = entries_map["subdir1"]
    assert subdir.is_directory()
    assert subdir.size == 512

    subdir_entries_map = {e.name: e for e in list(subdir.iterdir())}

    assert set(subdir_entries_map.keys()) == {".", "..", "file3.txt"}

    file3 = subdir_entries_map["file3.txt"]
    assert not file3.is_directory()
    assert file3.size == 27



from dissect.fat.exceptions import (
    BadClusterError,
    EmptyDirectoryError,
    Error,
    FileNotFoundError,
    FreeClusterError,
    InvalidBPB,
    InvalidDirectoryError,
    InvalidHeaderMagic,
    LastEmptyDirectoryError,
    NotADirectoryError,
)
from dissect.fat.exfat import ExFAT
from dissect.fat.fat import FATFS

__all__ = [
    "ExFAT",
    "FATFS",
    "BadClusterError",
    "EmptyDirectoryError",
    "Error",
    "FileNotFoundError",
    "FreeClusterError",
    "InvalidBPB",
    "InvalidDirectoryError",
    "InvalidHeaderMagic",
    "LastEmptyDirectoryError",
    "NotADirectoryError",
]



import logging
import os
import struct
from collections import OrderedDict
from itertools import groupby
from operator import itemgetter

from dissect.util.stream import RangeStream, RunlistStream

from dissect.fat.c_exfat import (
    BITMAP_ENTRY,
    DIR_ENTRY_SIZE,
    EOC,
    FAT_ENTRY_SIZE,
    FILE_ENTRY,
    NO_VOLUME_LABEL_ENTRY,
    UPCASE_TABLE_ENTRY,
    VOLUME_LABEL_ENTRY,
    c_exfat,
)
from dissect.fat.exceptions import InvalidHeaderMagic

__all__ = ["ExFAT"]

log = logging.getLogger(__name__)
log.setLevel(os.getenv("DISSECT_LOG_EXFAT", "CRITICAL"))


class ExFAT:
    def __init__(self, fh):
        self.filesystem = fh
        fh.seek(0)

        self.vbr = c_exfat.EXFAT_HEADER(fh.read(512))
        self.fs_name = self.vbr.fs_name
        if self.fs_name != b"EXFAT   ":
            raise InvalidHeaderMagic("Invalid exFAT header magic")

        # These get stored as exponents
        self.sector_size = 2**self.vbr.bytes_per_sector_exp
        self.cluster_size = self.sector_size * (2**self.vbr.sectors_per_cluster_exp)

        self.fat_sector = self.vbr.fat_sector  # Almost always sector 128 (absolute)
        self.fat_sector_count = self.vbr.fat_sector_count
        self.fat_size = self.fat_sector_count * self.sector_size
        self.fat = RangeStream(fh, (self.fat_sector * self.sector_size), self.fat_size)

        self.cluster_heap_sector = self.vbr.cluster_heap_sector
        self.cluster_count = self.vbr.cluster_heap_count
        self.root_dir_cluster = self.vbr.root_dir_cluster
        self.root_dir_sector = self.cluster_to_sector(self.root_dir_cluster)
        self.root_directory = RootDirectory(fh, self.root_dir_cluster, self)
        self.files = self.root_directory.dict  # maybe linked list to be able to traverse sub directories easier?

        self.alloc_bitmap = self.root_directory.bitmap_entry
        self.upcase_table = self.root_directory.upcase_entry
        self.volume_label = self.root_directory.volume_entry.volume_label.strip("\x00")

    def cluster_to_sector(self, cluster):
        """
        Returns the clusters' corresponding sector address

        Args:
            cluster (int): cluster address

        Returns:
            int: corresponding sector address if available
        """

        sector = ((cluster - 2) * (2**self.vbr.sectors_per_cluster_exp)) + self.cluster_heap_sector
        return sector if sector > 0 else None

    def sector_to_cluster(self, sector):
        """
        Returns the sectors' corresponding cluster address

        Args:
            sector (int): sector address

        Returns:
            int: corresponding cluster address if available
        """

        cluster = ((sector - self.cluster_to_sector(2)) // self.sector_size) + 2
        return cluster if cluster >= 2 else None

    def runlist(self, starting_cluster, not_fragmented=True, size=None):
        """
        Creates a RunlistStream compatible runlist from exFAT FAT structures

        Args:
            starting_cluster (int): First cluster of file, folder or location in question

        Returns:
            runlist: [(sector_offset, run_length)]"""

        # If file is not fragmented clusters will not be present in the FAT
        cluster_chain = [starting_cluster] if not_fragmented else self.get_cluster_chain(starting_cluster)
        runlist = []

        # TODO Graceful way to construct a runlist of non-fragmented streams spanning multiple sectors
        if size:
            run = -(-size // self.cluster_size)
            runlist.append((self.cluster_to_sector(cluster_chain[0]), run))
        else:
            # This is a somewhat convoluted, but short way to group successive
            # clusters together.
            # As the cluster numbers in the cluster_chain are strictly
            # incrementing, a succesive range of clusters will have the same
            # delta with respect to their position in the cluster_chain, which
            # is different from any other successive range, which is what is
            # used in the groupby().
            for _, cluster_group in groupby(enumerate(cluster_chain), lambda i: i[0] - i[1]):
                run = list(map(itemgetter(1), cluster_group))
                start_cluster = run[0]
                run_len = len(run)
                runlist.append((self.cluster_to_sector(start_cluster), run_len))

        return runlist

    def get_cluster_chain(self, starting_cluster):
        """
        Reads the on disk FAT to construct the cluster chain

        Args:
            starting_cluster (int): cluster to look-up the chain from

        Returns:
            list: Chain of clusters. Including starting_cluster
        """

        next_ = 0x00000000
        chain = []

        if starting_cluster < 2:
            return chain
        else:
            while next_ != EOC:
                self.fat.seek(starting_cluster * FAT_ENTRY_SIZE)
                next_ = struct.unpack("<L", self.fat.read(FAT_ENTRY_SIZE))[0]
                chain.append(starting_cluster)
                starting_cluster = next_

            return chain

    @staticmethod
    def _utc_timezone(timezone):
        """
        Converts a Microsoft exFAT timezone byte to its UTC timezone equivalent

        Args:
            timezone (int): exFAT timezone byte

        Returns:
            dict: UTC name (str), UTC offset in minutes (int)
        """

        #    utc_bool -64 32 16 : 8 4 2 1
        #              ^ sign
        TIMEZONE_MINUTE_INCREMENT = 15
        utc_enabled = timezone >> 7
        signed = (timezone >> 6) & 0x11  # Check if the second bit is set

        if utc_enabled:
            if signed:
                utc_minute_offset = ((timezone & 0x3F) - 64) * TIMEZONE_MINUTE_INCREMENT
            else:
                utc_minute_offset = (timezone & 0x7F) * TIMEZONE_MINUTE_INCREMENT

            hours, minutes = divmod(utc_minute_offset, 60)
            utc_name = f"UTC{hours:+03}:{minutes:02}"

            return {"name": utc_name, "offset": utc_minute_offset}
        else:
            return {"name": "localtime", "offset": 0}


class RootDirectory:
    def __init__(self, fh, location, exfat):
        self.exfat = exfat
        self.location = location
        self.size = 0
        self.root_dir = None
        self.volume_entry = None
        self.upcase_entry = None
        self.bitmap_entry = None
        self.dict = OrderedDict()
        self._parse_root_dir(fh)

    def _parse_root_dir(self, fh):
        """
        Parses the passed fh to construct the Root directory object"""

        # Root dir is always present in FAT so we pass False to traverse the FAT table
        # thus root dir is per definition fragmented
        runlist = self.exfat.runlist(self.location, False)
        size = 0

        # Calculate size of rootdir from runlist since rootdir has no size attribute
        for run in runlist:
            size += run[1] * self.exfat.cluster_size

        self.size = size
        self.root_dir = RunlistStream(fh, runlist, self.size, self.exfat.cluster_size)
        self.dict = self._create_root_dir(self.root_dir)

    def _parse_subdir(self, entry):
        """
        Parses the given sub directory file directory entry for containing files

        Args:
            entry (FILE): Directory FILE entry

        Returns:
            OrderedDict: Containing files in sub directory
        """

        folder_location = entry.stream.location
        folder_size = entry.stream.data_length
        folder_runlist = self.exfat.runlist(folder_location, not_fragmented=entry.stream.flags.not_fragmented)

        fh = RunlistStream(self.exfat.filesystem, folder_runlist, folder_size, self.exfat.cluster_size)
        return self._parse_file_entries(fh)

    @staticmethod
    def _construct_filename(fn_entries, is_dir=False):
        """
        Assembles the filename from given file name directory entries

        Args:
            fn_entries (list): A list of exFAT file name directory entries

        Returns:
            str: Name of file or folder stripped from trailing null values
        """

        filename = []
        if len(fn_entries) == 1:
            filename = fn_entries[0].filename.strip("\x00")
        else:
            for fn_entry in fn_entries:
                filename.append(fn_entry.filename)

            filename = "".join(filename).strip("\x00")

        return filename if not is_dir else filename + "/"

    def _parse_file_entries(self, fh):
        """
        Finds and parses file entries in a given file handle (file like object)

        Args:
            fh (Stream object): Any stream object

        Returns:
            OrderedDict: Found and parsed file directory entries
        """

        entries = OrderedDict()

        while fh.tell() < fh.size:
            entry = c_exfat.FILE_DIRECTORY_ENTRY(fh.read(DIR_ENTRY_SIZE))

            if entry.entry_type == FILE_ENTRY:  # Or entry.entry_type == 0x05:
                # Entry is a file so we reuse it
                metadata = entry

                # -1 because the metadata entry includes the stream dir entry in its count
                fnentry_count = metadata.subentry_count - 1

                stream = c_exfat.STREAM_DIRECTORY_ENTRY(fh.read(DIR_ENTRY_SIZE))
                fn_entries = []

                for _ in range(fnentry_count):
                    fn_entries.append(c_exfat.FILENAME_DIRECTORY_ENTRY(fh.read(DIR_ENTRY_SIZE)))

                file_ = c_exfat.FILE(metadata=metadata, stream=stream, fn_entries=fn_entries)
                if file_.metadata.attributes.directory:
                    # A directory will have its own file entry as it's first element
                    # and a ordered dict of file entry contained in it which can be accessed by their corresponding keys
                    filename = self._construct_filename(file_.fn_entries)
                    entries[filename] = (file_, self._parse_subdir(file_))
                else:
                    filename = self._construct_filename(file_.fn_entries)
                    entries[filename] = (file_, None)
            else:
                self._non_file_entries(entry)

        return entries

    def _non_file_entries(self, entry):
        if entry.entry_type == VOLUME_LABEL_ENTRY or entry.entry_type == NO_VOLUME_LABEL_ENTRY:
            self.volume_entry = c_exfat.VOLUME_DIRECTORY_ENTRY(entry.dumps())
        elif entry.entry_type == BITMAP_ENTRY:
            self.bitmap_entry = c_exfat.BITMAP_DIRECTORY_ENTRY(entry.dumps())
        elif entry.entry_type == UPCASE_TABLE_ENTRY:
            self.upcase_entry = c_exfat.UPCASE_DIRECTORY_ENTRY(entry.dumps())

    def _create_root_dir(self, root_dir):
        """
        Since exFAT does not have a dedicated root directory entry
        one has to be constructed form available parameters during filesystem parsing.

        This is ballpark so no real forensic conclusions should be based on the information of the root entry
        """

        metadata = c_exfat.FILE_DIRECTORY_ENTRY(
            entry_type=0x85, subentry_count=2, attributes=c_exfat.ATTRIBUTES(directory=0b1)
        )

        stream = c_exfat.STREAM_DIRECTORY_ENTRY(
            entry_type=0xC0,
            flags=c_exfat.FILE_FLAGS(allocation_possible=0b0, not_fragmented=0b0),
            filename_length=1,
            location=self.location,
            data_length=self.size,
        )

        fn_entries = [
            c_exfat.FILENAME_DIRECTORY_ENTRY(
                entry_type=0xC1, flags=0x00, filename="/\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            )
        ]

        root_entry = OrderedDict()
        root_entry["/"] = (
            c_exfat.FILE(metadata=metadata, stream=stream, fn_entries=fn_entries),
            self._parse_file_entries(root_dir),
        )

        return root_entry



# References:
# - https://ogris.de/fatrepair/fat.c
# - https://github.com/nathanhi/pyfatfs
# - https://download.microsoft.com/download/1/6/1/161ba512-40e2-4cc9-843a-923143f3456c/fatgen103.doc

import datetime
import struct
from functools import lru_cache, reduce
from operator import itemgetter

from dissect.util.stream import RangeStream, RunlistStream
from dissect.util.ts import dostimestamp

from dissect.fat.c_fat import (
    BAD_CLUSTER,
    DATA_CLUSTER_MAX,
    DATA_CLUSTER_MIN,
    END_OF_CLUSTER_MAX,
    END_OF_CLUSTER_MIN,
    FAT12_EOC,
    FREE_CLUSTER,
    VALID_BPB_MEDIA,
    Fattype,
    c_fat,
)
from dissect.fat.exceptions import (
    BadClusterError,
    EmptyDirectoryError,
    FileNotFoundError,
    FreeClusterError,
    InvalidBPB,
    InvalidDirectoryError,
    LastEmptyDirectoryError,
    NotADirectoryError,
)


class FATFS:
    def __init__(self, fh, encoding="ibm437"):
        self.fh = fh
        self.encoding = encoding

        sector = fh.read(512)
        bpb_size = len(c_fat.Bpb)
        self.bpb = c_fat.Bpb(sector[:bpb_size])
        bpb16 = c_fat.Bpb16(sector[bpb_size:])
        bpb32 = c_fat.Bpb32(sector[bpb_size:])

        validate_bpb(self.bpb)

        self.fat_size = self.bpb.BPB_FATSz16 or bpb32.BPB_FATSz32
        self.total_sectors = self.bpb.BPB_TotSec16 or self.bpb.BPB_TotSec32

        # Taken from FAT32 spec
        root_dir_sectors = ((self.bpb.BPB_RootEntCnt * 32) + (self.bpb.BPB_BytsPerSec - 1)) // self.bpb.BPB_BytsPerSec
        self.first_data_sector = self.bpb.BPB_RsvdSecCnt + (self.bpb.BPB_NumFATs * self.fat_size) + root_dir_sectors

        data_sec = self.total_sectors - self.first_data_sector
        count_of_clusters = data_sec // self.bpb.BPB_SecPerClus
        if count_of_clusters < 4085:
            self.type = Fattype.FAT12
            self.bpb_ext = bpb16
        elif count_of_clusters < 65525:
            self.type = Fattype.FAT16
            self.bpb_ext = bpb16
        else:
            self.type = Fattype.FAT32
            self.bpb_ext = bpb32

        self.sector_size = self.bpb.BPB_BytsPerSec
        self.cluster_size = self.bpb.BPB_BytsPerSec * self.bpb.BPB_SecPerClus

        # FAT starts after reserved sectors
        # Only parse the first FAT for now
        fat_stream = RangeStream(
            fh,
            self.bpb.BPB_RsvdSecCnt * self.sector_size,
            self.fat_size * self.sector_size,
        )
        self.fat = FAT(fat_stream, self.type)
        self.data_stream = RangeStream(fh, self.first_data_sector * self.sector_size, data_sec * self.sector_size)

        # volume label with stripped padding
        self.volume_label = bytes(self.bpb_ext.BS_VolLab).strip(b"\x20").decode(encoding)

        # volume serial number, hex encoded
        self.volume_id = f"{self.bpb_ext.BS_VolID:x}"

        self.root = RootDirectory(self)

    def get(self, path, dirent=None):
        dirent = self.root if not dirent else dirent

        # Programmatically we will often use the `/` separator, so replace it with the native path separator of FAT
        # `/` is an illegal character in FAT filenames, so it's safe to replace
        parts = path.replace("/", "\\").split("\\")
        for part in parts:
            if not part:
                continue

            if dirent is self.root and part in (".", ".."):
                continue

            part_upper = part.upper()
            for child in dirent.iterdir():
                if part_upper in (child.name.upper(), child.short_name.upper()):
                    dirent = child
                    break
            else:
                raise FileNotFoundError(f"File not found: {path}")

        return dirent


def validate_bpb(bpb):
    if isinstance(bpb, bytes):
        bpb = c_fat.Bpb(bpb[: len(c_fat.Bpb)])

    # Detect a valid x86 JMP opcode
    if not (bpb.BS_jmpBoot[0] == 0xEB and bpb.BS_jmpBoot[2] == 0x90) and bpb.BS_jmpBoot[0] != 0xE9:
        raise InvalidBPB(f"Invalid BS_jmpBoot: {bytes(bpb.BS_jmpBoot)!r}")

    if bpb.BPB_BytsPerSec not in [2**x for x in range(9, 13)]:
        raise InvalidBPB(f"Invalid BPB_BytsPerSec: 0x{bpb.BPB_BytsPerSec:x}")

    if bpb.BPB_SecPerClus not in [2**x for x in range(8)]:
        raise InvalidBPB(f"Invalid BPB_SecPerClus: 0x{bpb.BPB_SecPerClus:x}")

    if bpb.BPB_RsvdSecCnt == 0:
        raise InvalidBPB(f"Invalid BPB_RsvdSecCnt, must not be 0: 0x{bpb.BPB_RsvdSecCnt:x}")

    if bpb.BPB_Media not in VALID_BPB_MEDIA:
        raise InvalidBPB(f"Invalid BPB_Media: 0x{bpb.BPB_Media:x}")

    if bpb.BPB_NumFATs < 1:
        raise InvalidBPB(f"Invalid BPB_NumFATs, must be at least 1: 0x{bpb.BPB_NumFATs:x}")

    root_entry_count = bpb.BPB_RootEntCnt * 32
    root_entry_count %= bpb.BPB_BytsPerSec
    if bpb.BPB_RootEntCnt != 0 and root_entry_count != 0:
        raise InvalidBPB("Root entry count does not align with bytes per sector")

    if bpb.BPB_TotSec16 == 0 and bpb.BPB_TotSec32 == 0:
        raise InvalidBPB(f"Invalid BPB_TotSec16 and BPB_TotSec32: 0x{bpb.BPB_TotSec16:x}, 0x{bpb.BPB_TotSec32:x}")


class FAT:
    def __init__(self, fh, fattype):
        self.fh = fh

        if fattype == Fattype.FAT12:
            self.bits_per_entry = 12
        elif fattype == Fattype.FAT16:
            self.bits_per_entry = 16
        elif fattype == Fattype.FAT32:
            self.bits_per_entry = 32
        else:
            raise TypeError("Unsupported FAT type")

        self.entry_count = int(self.fh.size // (self.bits_per_entry / 8))

    @lru_cache(4096)
    def get(self, cluster):
        if cluster >= self.entry_count:
            raise ValueError(f"Cluster exceeds FAT entry count: {cluster} >= {self.entry_count}")

        if self.bits_per_entry == 12:
            offset_in_fat = cluster + (cluster // 2)
            self.fh.seek(offset_in_fat)
            value = struct.unpack("<H", self.fh.read(2))[0]

            if cluster & 1:
                return value >> 4
            else:
                return value & 0x0FFF
        elif self.bits_per_entry == 16:
            offset_in_fat = cluster * 2
            self.fh.seek(offset_in_fat)
            return struct.unpack("<H", self.fh.read(2))[0]
        elif self.bits_per_entry == 32:
            offset_in_fat = cluster * 4
            self.fh.seek(offset_in_fat)
            return struct.unpack("<I", self.fh.read(4))[0] & 0x0FFFFFFF  # FAT32 clusters are 28 bits

    def chain(self, cluster):
        bits = self.bits_per_entry
        while True:
            value = self.get(cluster)

            if DATA_CLUSTER_MIN <= value <= mask(DATA_CLUSTER_MAX, bits):
                yield cluster

            # FAT12 special EOC
            if self.bits_per_entry == 12 and value == FAT12_EOC:
                yield cluster
                break

            if mask(END_OF_CLUSTER_MIN, bits) <= value <= mask(END_OF_CLUSTER_MAX, bits):
                yield cluster
                break

            if value == mask(BAD_CLUSTER, bits):
                raise BadClusterError(cluster)

            if value == FREE_CLUSTER:
                raise FreeClusterError(cluster)

            cluster = value

    def runlist(self, cluster):
        """Create a runlist from a cluster chain.

        First two clusters are reserved, so substract those.
        Also combine consecutive clusters for a more efficient runlist.
        """
        chain = self.chain(cluster)

        run_start = next(chain) - 2
        run_size = 1

        for cl in chain:
            if cl == run_start + run_size:
                run_size += 1
            else:
                yield (run_start, run_size)
                run_start = cl - 2
                run_size = 1
        else:
            yield (run_start, run_size)


class DirectoryEntry:
    def __init__(self, fs, fh, parent=None):
        self.fs = fs
        self.parent = parent

        self.dirent = None
        self.ldirents = []

        self.name = None
        self.short_name = None

        buf = fh.read(32)
        dirent = c_fat.Dirent(buf)

        if dirent.DIR_Name[0] == 0xE5:
            raise EmptyDirectoryError("Dirent is an empty entry")

        if dirent.DIR_Name[0] == 0x0:
            raise LastEmptyDirectoryError("Dirent is the last empty entry")

        if dirent.DIR_Attr & c_fat.ATTR_LONG_NAME_MASK == c_fat.ATTR_LONG_NAME:
            ldirent = c_fat.Ldirent(buf)
            found_last = False
            while ldirent.LDIR_Attr == c_fat.ATTR_LONG_NAME:
                self.ldirents.append(ldirent)
                if ldirent.LDIR_Ord & c_fat.LAST_LONG_ENTRY:
                    if found_last:
                        raise InvalidDirectoryError("Dirent contains multiple last-long entries")
                    found_last = True

                buf = fh.read(32)
                ldirent = c_fat.Ldirent(buf)

            self.dirent = c_fat.Dirent(buf)

            self.ldirents.sort(key=lambda e: e.LDIR_Ord & 0x3F)
            name_map = map(lambda e: e.LDIR_Name1 + e.LDIR_Name2 + e.LDIR_Name3, self.ldirents)
            name_reduce = bytes(reduce(lambda a, b: a + b, name_map))
            self.name = c_fat.wchar[None](name_reduce + b"\x00\x00")
        else:
            self.dirent = dirent

        dir_name = bytearray(self.dirent.DIR_Name)
        if dir_name[0] == 0x05:
            dir_name[0] = 0xE5

        base = dir_name[:8].decode(self.fs.encoding).rstrip("\x00").rstrip()
        ext = dir_name[8:].decode(self.fs.encoding).rstrip("\x00").rstrip()

        self.short_name = ".".join([base, ext]) if ext else base
        if not self.name:
            self.name = self.short_name

        self._runlist = None
        self._entries = None

    def __repr__(self):
        return f"<DirectoryEntry name={self.name}>"

    @property
    def path(self):
        return "\\".join([self.parent.path if self.parent else "", self.name]).lstrip("\\")

    @property
    def size(self):
        if self.is_directory():
            return sum(map(itemgetter(1), self.dataruns())) * self.fs.cluster_size
        else:
            return self.dirent.DIR_FileSize

    @property
    def cluster(self):
        return (self.dirent.DIR_FstClusHI << 16) | self.dirent.DIR_FstClusLO

    @property
    def ctime(self):
        if self.dirent.DIR_CrtDate and self.dirent.DIR_CrtTime:
            return dostimestamp(
                (self.dirent.DIR_CrtDate << 16) | self.dirent.DIR_CrtTime,
                self.dirent.DIR_CrtTimeTenth,
            )
        return datetime.datetime(1980, 1, 1)

    @property
    def atime(self):
        if self.dirent.DIR_LstAccDate:
            return dostimestamp(self.dirent.DIR_LstAccDate << 16)
        return datetime.datetime(1980, 1, 1)

    @property
    def mtime(self):
        return dostimestamp((self.dirent.DIR_WrtDate << 16) | self.dirent.DIR_WrtTime)

    def is_readonly(self):
        return bool(self.dirent.DIR_Attr & c_fat.ATTR_READ_ONLY)

    def is_hidden(self):
        return bool(self.dirent.DIR_Attr & c_fat.ATTR_HIDDEN)

    def is_system(self):
        return bool(self.dirent.DIR_Attr & c_fat.ATTR_SYSTEM)

    def is_volume_id(self):
        return bool(self.dirent.DIR_Attr & c_fat.ATTR_VOLUME_ID)

    def is_directory(self):
        return bool(self.dirent.DIR_Attr & c_fat.ATTR_DIRECTORY)

    def is_archive(self):
        return bool(self.dirent.DIR_Attr & c_fat.ATTR_ARCHIVE)

    def iterdir(self):
        if not self.is_directory():
            raise NotADirectoryError(self.name)

        if not self._entries:
            entries = []
            for entry in _iter_dirent(self.fs, self.open(), self):
                yield entry
                entries.append(entry)

            self._entries = entries
        else:
            yield from self._entries

    def dataruns(self):
        if self._runlist is None:
            self._runlist = [] if self.cluster == FREE_CLUSTER else list(self.fs.fat.runlist(self.cluster))
        return self._runlist

    def open(self):
        return RunlistStream(self.fs.data_stream, self.dataruns(), self.size, self.fs.cluster_size)


class RootDirectory(DirectoryEntry):
    def __init__(self, fs):
        self.fs = fs
        self.name = "\\"
        self.short_name = self.name

        self._runlist = None
        self._entries = None

    @property
    def path(self):
        return ""

    @property
    def size(self):
        if self.fs.type in (Fattype.FAT12, Fattype.FAT16):
            return self.fs.bpb.BPB_RootEntCnt * 32
        else:
            return sum(map(itemgetter(1), self.dataruns())) * self.fs.cluster_size

    @property
    def cluster(self):
        if self.fs.type == Fattype.FAT32:
            return self.fs.bpb_ext.BPB_RootClus
        return None

    @property
    def ctime(self):
        return datetime.datetime(1980, 1, 1)

    @property
    def atime(self):
        return datetime.datetime(1980, 1, 1)

    @property
    def mtime(self):
        return datetime.datetime(1980, 1, 1)

    def is_readonly(self):
        return False

    def is_hidden(self):
        return False

    def is_system(self):
        return False

    def is_volume_id(self):
        return False

    def is_directory(self):
        return True

    def is_archive(self):
        return False

    def iterdir(self):
        if not self._entries:
            entries = []
            for entry in _iter_dirent(self.fs, self.open(), self):
                yield entry
                entries.append(entry)

            self._entries = entries
        else:
            yield from self._entries

    def dataruns(self):
        if self._runlist is None:
            self._runlist = [] if self.cluster == FREE_CLUSTER else list(self.fs.fat.runlist(self.cluster))
        return self._runlist

    def open(self):
        if self.fs.type in (Fattype.FAT12, Fattype.FAT16):
            root_dir_sector = self.fs.bpb.BPB_RsvdSecCnt + (self.fs.fat_size * self.fs.bpb.BPB_NumFATs)
            offset = root_dir_sector * self.fs.sector_size
            return RangeStream(self.fs.fh, offset, self.size)
        else:
            return RunlistStream(self.fs.data_stream, self.dataruns(), self.size, self.fs.cluster_size)


def mask(v, bits):
    return v & ((1 << bits) - 1)


def _iter_dirent(fs, fh, parent=None):
    while True:
        try:
            yield DirectoryEntry(fs, fh, parent)
        except EmptyDirectoryError:
            continue
        except LastEmptyDirectoryError:
            break
        except EOFError:
            break



from dissect import cstruct

c_exfat_def = """
enum entry_types : uint8 {
    allocation_bitmap   = 0x81,
    upcase_table        = 0x82,
    volume_label        = 0x83,
    file_directory      = 0x85,
    stream_directory    = 0xC0,
    filename_directory  = 0xC1
};

typedef struct TIMESTAMP {
    uint32 seconds:5;
    uint32 minutes:6;
    uint32 hours:5;
    uint32 day:5;
    uint32 month:4;
    uint32 year:7;
};

typedef struct ENTRY_TYPE {
    uint8 in_use:1;
    uint8 category:1;
    uint8 importance:1;
    uint8 type_code:5;
};

typedef struct ATTRIBUTES {
    uint16 read_only:1;
    uint16 hidden:1;
    uint16 system:1;
    uint16 reserved_1:1;
    uint16 directory:1;
    uint16 archive:1;
    uint16 reserved:10;
};

typedef struct FILE_FLAGS {
    uint8 allocation_possible:1;
    uint8 not_fragmented:1;
    uint8 reserved_1:6;
};

typedef struct EXFAT_HEADER {
    uint8   jump_boot[3];               // 0x00 jmp and nop instructions
    char    fs_name[8];			        // 0x03 "EXFAT   "
    uint8   fat32[53];			        // 0x0B always 0 to avoid accidental mounting as FAT32
    uint64  volume_offset;			    // 0x40 first sector of exFAT volume
    uint64  volume_sector_count;	    // 0x48 sectors count of whole volume
    uint32  fat_sector;		            // 0x50 FAT first sector
    uint32  fat_sector_count;		    // 0x54 FAT sectors count
    uint32  cluster_heap_sector;	    // 0x58 first cluster sector
    uint32  cluster_heap_count;			// 0x5C total clusters count
    uint32  root_dir_cluster;			// 0x60 first cluster of the root dir
    uint32  volume_serial;			    // 0x64 volume serial number
    uint8   fs_version_min;             // 0x68 minor portion of FS version
    uint8   fs_version_maj;             // 0x69 major portion of FS version
    uint16  volume_flags;		    	// 0x6A volume state flags
    uint8   bytes_per_sector_exp;	    // 0x6C sector size as (1 << n)
    uint8   sectors_per_cluster_exp;	// 0x6D sectors per cluster as (1 << n)
    uint8   number_of_fats;  		    // 0x6E always 1
    uint8   drive_select;			    // 0x6F always 0x80
    uint8   percent_inuse;		        // 0x70 percentage of allocated space
    uint8   zeros_2[7];                 // 0x71 always 0
    uint8   boot_code[390];			    // 0x78 code containing bootloader for non-*nix systems
    uint16  boot_signature; 			// 0x1FE boot sector signature 0xAA55
};

typedef struct STREAM_DIRECTORY_ENTRY {
    uint8 entry_type;                  // 0x00 type of directory entry should be 0xC0 0x40
    FILE_FLAGS flags;                  // 0x01 only first 2 bits get used and only bit 1 is relevant True equals
                                       // notfragmented
    uint8 reserved_1;                  // 0x02 actually part of is_fragmented but does not get used in
                                       // current exFAT version
    uint8 filename_length;             // 0x03 length of filename In filename directory entry
    uint16 filename_hash;              // 0x04 hash of filename used to speed up lookups within exFAT
    uint16 protected;                  // 0x06 indicates whether file is protected by EFS
    uint64 valid_data_length;          // 0x08 allocated size of data in bytes only counts for files and is used for
                                       // pre-allocation zero if directory
    uint8 reserved_2[4];               // 0x10 no clue always zero
    uint32 location;                   // 0x15 starting cluster of data
    uint64 data_length;                // 0x18 actual size of data if directory always multiples of sector size
};

typedef struct FILENAME_DIRECTORY_ENTRY {
    uint8 entry_type;                  // 0x00 type of directory entry should be 0xC1 or 0x41
    uint8 flags;                       // 0x01 always zero for filename directory entry
    wchar filename[15];                // 0x02 filename of dir or file each char is represented as 16 bit unicode
};

typedef struct FILE_DIRECTORY_ENTRY {
    uint8  entry_type;                 // 0x00 type of directory entry should be 0x85 or 0x05
    uint8  subentry_count;             // 0x01 amount of secondary directory entries (minimal 2) stream and filename
    uint16 checksum;                   // 0x02 checksum off whole set File dir entries and subentry_count
    ATTRIBUTES attributes;             // 0x04 Microsoft file attrib exFAT uses only the first 5 bits
    uint8  zeros_1[2];                 // 0x06 always zero. checksum gets calculated over it
    uint32 create_time;                // 0x08 MS-DOS Timestamp format for created time
    uint32 modified_time;              // 0x0C MS-DOS Timestamp format for last modified time
    uint32 access_time;                // 0x10 MS-DOS Timestamp format for last access time
    uint8  create_offset;              // 0x14 centisecond offset of create time
    uint8  modified_offset;            // 0x15 centisecond offset of modified time
    uint8  create_timezone;            // 0x16 timezone offset to UTC of Create time
    uint8  modified_timezone;          // 0x17 timezone offset to UTC of Modified time
    uint8  access_timezone;            // 0x18 timezone offset to UTC of Access time
    uint8  zeros_2[7];                 // 0x19 always zero. Checksum gets calculated over it
};

typedef struct FILE {
    FILE_DIRECTORY_ENTRY     metadata; // collection of entries that define a file or directory
    STREAM_DIRECTORY_ENTRY   stream;
    FILENAME_DIRECTORY_ENTRY fn_entries[];
};

typedef struct VOLUME_DIRECTORY_ENTRY {
    uint8 entry_type;                  // 0x00 type of directory entry should be 0x83 or 0x03
    uint8 label_length;                // 0x01 length of volume label string
    wchar volume_label[11];            // 0x02 volume label string max 11 chars
    uint8 zeros_1[8];                  // 0x17 no clue always zero maybe volume label overflow chars
};

typedef struct BITMAP_DIRECTORY_ENTRY {
    uint8 entry_type;                 // 0x00 type of directory entry should be 0x81
    uint8 flags;                      // 0x01 indicates how many bitmaps are used always 0
                                      // which indicates 1 bitmaps is used
    uint8 reserved_1[18];             // 0x02 always zero reserved for later use
    uint32 bitmap_start_cluster;      // 0x0 cluster addr of the bitmap
    uint64 bitmap_length;             // 0x0 length of bitmap in bytes
};

typedef struct UPCASE_DIRECTORY_ENTRY {
    uint8 entry_type;                 // 0x00 type of directory entry should be 0x82
    uint8 reserved_1[3];              // 0x02 always zero reserved for later use
    uint32 upcase_checksum;           // 0x04 checksum of upcase tabel should always be 0DD319E6
    uint8 reserved_2[12];             // 0x08 always zero
    uint32 upcase_start_cluster;      // 0x0C cluster addr of upcase tabel
    uint64 upcase_length;             // 0x12 length of upcase tabel in bytes
};
"""

# default endianess is LE so we keep it that way.
c_exfat = cstruct.cstruct()
c_exfat.load(c_exfat_def)

EOC = 0xFFFFFFFF  # indicates end of cluster chain
FID = 0xFFFFFFF8  # indicates start of FAT

VOLUME_LABEL_ENTRY = c_exfat.entry_types.volume_label.value
FILE_ENTRY = c_exfat.entry_types.file_directory.value
FILENAME_ENTRY = c_exfat.entry_types.filename_directory.value
UPCASE_TABLE_ENTRY = c_exfat.entry_types.upcase_table.value
STREAM_ENTRY = c_exfat.entry_types.stream_directory.value
BITMAP_ENTRY = c_exfat.entry_types.allocation_bitmap.value

NO_VOLUME_LABEL_ENTRY = 0x03
FAT_ENTRY_SIZE = 4
DIR_ENTRY_SIZE = 32



from dissect.fat import exfat


def test_exfat(exfat_simple):
    e = exfat.ExFAT(exfat_simple)

    assert e.volume_label == "THESIS"
    assert e.cluster_count == 1792
    assert e.sector_size == e.cluster_size
    assert e.fat_sector == 128
    assert e.root_dir_cluster == 15
    assert e.root_dir_sector == 269
    assert e.runlist(e.root_dir_cluster) == [(e.root_dir_sector, 1)]

    files = e.files
    assert sorted(files.keys()) == ["/"]

    root = files["/"][0]
    assert root.metadata.attributes.directory == 1
    assert root.stream.flags.not_fragmented == 0
    assert root.stream.data_length == 512

    cat = files["/"][1]["cat.jpg"][0]
    assert cat.metadata.attributes.directory == 0
    assert cat.stream.flags.not_fragmented == 1
    assert cat.stream.data_length == 88786

    find_me = files["/"][1]["find_me.txt"][0]
    assert find_me.metadata.attributes.directory == 0
    assert find_me.stream.flags.not_fragmented == 1
    assert find_me.stream.data_length == 9

    directory = files["/"][1]["directory"][0]
    assert directory.metadata.attributes.directory == 1
    assert directory.stream.flags.not_fragmented == 1
    assert directory.stream.data_length == 512

    sysvol = files["/"][1]["System Volume Information"][0]
    assert sysvol.metadata.attributes.directory == 1
    assert sysvol.stream.flags.not_fragmented == 1
    assert sysvol.stream.data_length == 512
